/*
	ProWesS global variables support

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 07 January 1996
	26.04.1996 added delete and iterator access function
			   duplicates are now removed
	22.06.1996 now safe cfr deletes etc (because values are copied instead of
			   passing back a pointer)
			   The length of the name is now limited to 64 chars
*/

#include "sms_h"
#include "mem_h"
#include "str_h"
#include "atomic_h"
#include "global_h"

#define catch(x) if (err=(x)) return err; else

#define MAX_NAME	64

typedef struct _Var {
	struct _Var *next, *prev;		/* next variable in list */
	char *name; 					/* name */
	char value[2];					/* value */
} Var;

Var head,last;

Error GLOBALInit()
{
	head.next=head.prev=&last;
	last.next=last.prev=&head;
	return ERR_OK;
}

static Error get(char *name, char *string, Size size)
{
	Var *curr=head.next;
	for ( ; curr!=&last ; curr=curr->next)
	{
		if (STRSameCD(curr->name,name))
		{
			STRCopySafe(curr->value,string,size);
			return ERR_OK;
		}
	}
	return ERR_ITNF;
}

Error GLOBALGet(char *name, char *string, Size size)
{
	return ATOMICCall(get,3,name,string,size);
}

static Error delete(char *name)
{
	Var *curr=head.next;
	for ( ; curr!=&last ; curr=curr->next)
	{
		if (STRSameCD(curr->name,name))
		{
			curr->next->prev=curr->prev;
			curr->prev->next=curr->next;
			return ERR_OK;
		}
	}
	return ERR_OK;
}

Error GLOBALDelete(char *name)
{
	return ATOMICCall(delete,1,name);
}

static Error link(Var *var)
{
	delete(var->name);
	var->prev=&head;
	var->next=head.next;
	head.next->prev=var;
	head.next=var;
	return ERR_OK;
}

Error GLOBALSet(char *name, char *value)
{
	Error err;
	Var *var;

	catch( MEMAllocateFor(strlen(value)+sizeof(Var)+min(strlen(name),MAX_NAME),
						  SYSTEM, &var) );
	var->name=var->value+strlen(value)+1;
	STRCopy(value,var->value);
	STRCopySafe(name,var->name,MAX_NAME);
	return ATOMICCall(link,1,var);
}

static Error first(char *name, char *string, Size size)
{
	if (head.next==&last) return ERR_EOF;
	STRCopy(head.next->name,name);
	STRCopySafe(head.next->value,string,size);
	return ERR_OK;
}

Error GLOBALFirst(char *name, char *string, Size size)
{
	return ATOMICCall(first,3,name,string,size);
}

static Error next(char *name, char *string, Size size)
{
	Var *curr=head.next->next;

	if (curr==&head) return ERR_EOF;

	for ( ; curr!=&last ; curr=curr->next)
	{
		if (STRSameCD(name,curr->prev->name))
		{
			STRCopy(curr->name,name);
			STRCopySafe(curr->value,string,size);
			return ERR_OK;
		}
	}
	return ERR_EOF;
}

Error GLOBALNext(char *name, char *string, Size size)
{
	return ATOMICCall(next,3,name,string,size);
}
